# Copyright 2001 by Tarjei Mikkelsen. All rights reserved.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""BioPython Pathway support module.

Bio.Pathway.Rep is a collection of general data abstractions used in the
implementation of the Bio.Pathway module. These abstractions are not intended
for direct use, but if they fit your needs there's no reason to reinvent the
wheel.
"""
